#include <stdio.h>
#include <stdlib.h>
#include <string.h>




static unsigned int get_word(unsigned char *p){
	return (p[0] | (p[1] << 8));
}

static unsigned long get_dword(unsigned char *p){
	return (p[0] | (p[1] << 8) | (p[2] << 16) | (p[3] << 24));
}



static int expand(unsigned char *dest, unsigned char *src, long src_size){
	unsigned char tbl1[256], tbl2[256], stack[128];
	long ofs, dest_ofs;
	int c, len;
	int stack_pos;

	memset(tbl1, 0x00, sizeof(tbl1));
	memset(tbl2, 0x00, sizeof(tbl2));
	memset(stack, 0x00, sizeof(stack));

	ofs = 4;
	dest_ofs = 0;

	while(ofs < src_size){
		for(c = 0; 0x100 > c; ++c){
			tbl1[c] = c;
		}

		len = src[ofs];
		ofs += 1;

		c = 0;
		while(1){
			if(0x7f < len){
				len -= 0x7f;
				c += len;
				len = 0;
			}
			if(0x100 == c){
				break;
			}
			len += 1;

			for(; 0 < len; --len){
				tbl1[c] = src[ofs];
				ofs += 1;
				if(tbl1[c] != c){
					tbl2[c] = src[ofs];
					ofs += 1;
				}
				c += 1;
			}

			if(0x100 == c){
				break;
			}
			len = src[ofs];
			ofs += 1;
		}

		len = ((src[ofs] << 8) | src[ofs + 1]);
		ofs += 2;

		stack_pos = 0;

		while(1){
			if(0 != stack_pos){
				stack_pos -= 1;
				c = stack[stack_pos];
			}
			else{
				if(0 >= len){
					break;
				}
				len -= 1;
				c = src[ofs];
				ofs += 1;
			}

			if(c == tbl1[c]){
				dest[dest_ofs] = c;
				dest_ofs += 1;
				continue;
			}

			stack[stack_pos] = tbl2[c];
			stack_pos += 1;
			stack[stack_pos] = tbl1[c];
			stack_pos += 1;
		}

	}

	return dest_ofs;
}



static long exp_size;

static void* x_expand(FILE *fi, long ofs, long size){
	void *src, *dest;
	long dest_size;

	fseek(fi, ofs, SEEK_SET);
	src = malloc(size);
	if(NULL == src){
		return NULL;
	}
	fread(src, 1, size, fi);
	dest_size = get_dword(src);
	dest = malloc(dest_size);
	if(NULL == dest){
		free(src);
		return NULL;
	}
	dest_size = expand(dest, src, size);
	exp_size = dest_size;
	free(src);
	return dest;
}


static void* x_read(FILE *fi, long ofs, long size){
	void *src;
	fseek(fi, ofs, SEEK_SET);
	src = malloc(size);
	if(NULL == src){
		return NULL;
	}
	fread(src, 1, size, fi);
	return src;
}




int main(int argc, char *argv[]){
	long src_size;
	unsigned char *tbl;
	long tbl_limit;
	FILE *fi;
	int flag;
	char *src_name;


	if(2 > argc){
		return 0;
	}

	flag = 0;
	src_name = NULL;
	{
		int k;
		for(k = 1; argc > k; ++k){
			if((argv[k][0] == '-') || (argv[k][0] == '/')){
				switch(argv[k][1]){
					case 'M':
					case 'm':
						flag = 1;
						break;
				}
			}
			else if(NULL == src_name){
				src_name = argv[k];
			}
		}
	}

	if(NULL == src_name){
		return 0;
	}


	{
		unsigned char buf[16];
		long ofs, size;

		fi = fopen(src_name, "rb");
		if(NULL == fi){
			return 1;
		}

		fread(buf, 1, 16, fi);
		fseek(fi, 0, SEEK_END);
		src_size = ftell(fi);

		if(0 != memcmp(buf, "AR\0\0", 4)){
			fclose(fi);
			return 1;
		}

		ofs = get_dword(buf + 0x0a);
		size = (src_size - ofs);

		tbl = x_expand(fi, ofs, size);
		if(NULL == tbl){
			fclose(fi);
			return 1;
		}
		tbl_limit = get_word(buf + 0x0e);
	}


	{
		long k;
		long ofs;

		ofs = 0;
		for(k = 0; ofs < tbl_limit; ++k){
			long t_ofs, t_size;
			void *buf;

			t_ofs = get_dword(tbl + ofs);
			t_size = get_word(tbl + ofs + 8);
			t_ofs &= 0xffffff;


			if((0 != flag) && (tbl[ofs + 7] != 0x17) && (tbl[ofs + 7] != 0x18)){
				ofs += 12;
				continue;
			}

			if(0 != tbl[ofs + 4]){
				buf = x_expand(fi, t_ofs, t_size);
				t_size = exp_size;
			}
			else{
				buf = x_read(fi, t_ofs, t_size);
			}
			if(NULL != buf){
				FILE *fo;
				char name[16];
				sprintf(name, "%04d.dat", k);
				fo = fopen(name, "wb");
				fwrite(buf, 1, t_size, fo);
				fclose(fo);
				free(buf);
			}

			ofs += 12;
		}
	}

	free(tbl);
	fclose(fi);

	return 0;
}
